#include <stdio.h>

/* Config interrupt rate */

main()
{
	struct foo
	{
		unsigned char rx_chars;
		unsigned char tx_chars;
		unsigned short int_count;
	} hw ;

	int fd;
	int int_count;

	fd=open("/dev/sitty0", 0);

	for(;;)
	{
		printf("Enter new int_count : ");
		scanf("%d", &int_count);

		hw.int_count = int_count;

		if(ioctl(fd, 0x5498, &hw)!=0) /* SET */
		{
			printf("\n\n\n");
			perror("test failed");
			printf("\n\n\n");
			exit(0);
		}


	}


}
